"""Describes an instance of our problem.
"""

from pyhrtc.basics import Agent, Instance


class WeightedAgent(Agent):
    """An agent in an instance of SMTI-GRP.
    """

    def __init__(self, ident, capacity=1):
        super().__init__(ident, capacity=capacity)
        self._preference_weights = []
        self._weights = {}
        self._sorted_preferences = None

    def add_weight(self, other_ident, weight):
        """Adds the given weight for the given ident.
        :param other_ident: The ID of the other agent
        :type other_ident: integer
        :param weight: The weight of other_ident according to this agent
        :type weight: float
        """
        self._preference_weights.append((other_ident, weight))
        self._weights[other_ident] = weight

    @property
    def preferences(self):
        """Return the preferences, as a list of tie groups. Entries which are
        not in a tie at all still appear by themselvs in a list (i.e. 1 (2 3)
        is [[1], [2, 3]]).
        """
        self._sort_preferences()
        return self._preferences

    def weight_of(self, other):
        """Return the weight of matching self with the agent "other", or raises
        an exception if this agent is not compatible with other.
        :param other: The ID of the agent whose weight you want
        :type other: integer
        :return: the weight of matching self with other
        :rtype: float
        """
        return self._weights[other]

    def threshold(self, threshold):
        """Remove any scores below the given threshold.
        :param threshold: The value under which no scores will be considered
        :type threshold: integer
        """
        self._sort_preferences()
        while (self._sorted_preferences and
               self._sorted_preferences[-1][1] < threshold):
            ident, _ = self._sorted_preferences.pop()
            del self._weights[ident]
        self._num_preferences = None
        self._build_preferences()

    def better_than(self, ident):
        """Returns the list of WeightedAgent IDs that this agent prefers as good as, or
        more than the one given by ident.
        :param ident: An ID for an WeightedAgent to compare to.
        :type ident: integer
        """
        self._sort_preferences()
        result = []
        agent_score = None
        for pref in self._sorted_preferences:
            if pref[0] == ident:
                agent_score = pref[1]
            if agent_score is not None and pref[1] < agent_score:
                return result
            result.append(pref[0])
        return result

    def preference_string(self):
        """Returns the string of preferences for this agent."""
        def format_agent(ident):
            """Show an agent as their id and weight."""
            return "%d (%f)" % (ident, self._weights[ident])

        def format_tie(tie_as_list):
            """Given a set of tied elements, returns a string representing
            them, by surrounding with brackets if there is more than one item.
            """
            if not tie_as_list:
                return ""
            if len(tie_as_list) == 1:
                return format_agent(tie_as_list[0])
            return "(%s)" % (" ".join([format_agent(ident)
                                       for ident in tie_as_list]))
        return " ".join([format_tie(tie) for tie in self.preferences])

    def _sort_preferences(self):
        """Sorts the preferences by score, highest to lowest.
        """
        # Do nothing if the list is already sorted.
        if self._sorted_preferences is not None:
            return
        self._sorted_preferences = sorted(self._preference_weights,
                                          reverse=True,
                                          key=lambda x: x[1])
        self._build_preferences()

    def _build_preferences(self):
        """Builds up the preferences structure from the _sorted_preferences
        structure.
        """
        last_weight = None
        self._preferences = []
        group = []
        for ident, weight in self._sorted_preferences:
            if last_weight is not None and weight != last_weight:
                self._preferences.append(group)
                group = []
            group.append(ident)
            last_weight = weight
        if group:
            self._preferences.append(group)

    def __str__(self):
        """A human readable string representation of this Agent.
        """
        return (f"WeightedAgent %d with preferences: %s" %
                (self._ident, self.preference_string()))


class WeightedInstance(Instance):
    """Represents an instance of SMTI-GRP. We assume that both sets of agents
    all have all weights assigned.

    :param lefts: One set of WeightedAgent objects
    :type lefts: Dict[:class:`WeightedAgent`]
    :param rights: The other set of WeightedAgent objects
    :type rightss: Dict[:class:`WeightedAgent`]
    """
    def __init__(self, lefts, rights):
        super().__init__(single_agents_left=lefts, single_agents_right=rights)

    def threshold(self, threshold):
        """Remove any scores below the given threshold from any agents in this
        instance.
        :param threshold: The value under which no scores will be considered
        :type threshold: integer
        """
        to_remove = []
        for left in self._single_agents_left.values():
            left.threshold(threshold)
            if left.is_empty():
                to_remove.append(left.ident)
        for ident in to_remove:
            del self._single_agents_left[ident]
        to_remove = []
        for right in self._single_agents_right.values():
            right.threshold(threshold)
            if right.is_empty():
                to_remove.append(right.ident)
        for ident in to_remove:
            del self._single_agents_right[ident]

    def weight(self, matching):
        """Get the weight of the given matching.
        :param matching: a dictionary mapping left_index to right_index such
        that Agent number left_index on the left is matching with Agent
        right_index on the right.
        """
        return sum(self._single_agents_left[l].weight_of(r)
                   for l, r in matching.items())

    @property
    def agents_left(self):
        """Returns a list of the WeightedAgents on the left side of this SMTI instance.
        :return: a list of WeightedAgents on the left side
        :rtype: List[:class:`WeightedAgent`]
        """
        return self._single_agents_left

    @property
    def agents_right(self):
        """Returns a list of the WeightedAgents on the right side of this SMTI
        instance.
        :return: a list of WeightedAgents on the right side
        :rtype: List[:class:`WeightedAgent`]
        """
        return self.number_of_single_agents_right

    def __str__(self):
        """A human readable string representation of this instance.
        """
        return (f"WeightedInstance with %d single agents on the left, %d "
                 "couples on the left and %d single agents on the right" %
                (self.number_of_single_agents_left(),
                 self.number_of_couples_left(),
                 self.number_of_single_agents_right()))



"""Encode an instance of SMTI as SAT in CNF, potentially with weights.
"""


def create_sat_smti(instance):
    """Creates a CNF SAT encoding of an SMTI instance.
    :raises NotImplementedError: Raised if the instance is not an SMTI instance
    """
    if not instance.is_SMTI():
        raise NotImplementedError

    varmap = {}
    var_count = 1
    for agent in instance.single_residents:
        pref_length = 1
        for pref_group in agent.preferences:
            for pref in pref_group:
                varmap[("r", agent.id, pref_length)] = var_count
                var_count += 1
                pref_length += 1
        varmap[("r", agent.id, pref_length)] = var_count
        var_count += 1
    for agent in instance.single_residents:
        pref_length = 1
        for pref_group in agent.preferences:
            for pref in pref_group:
                varmap[("h", agent.id, pref_length)] = var_count
                var_count += 1
                pref_length += 1
        varmap[("h", agent.id, pref_length)] = var_count
        var_count += 1

    encoding = ""
    # Clause 1
    for resident in instance.single_residents:
        encoding += "%d 0" % varmap[("r", resident.id, 1)]
    # Clause 2
    for hospital in instance.hospitals:
        encoding += "%d 0" % varmap[("h", hospital.id, 1)]
    # Clause 3
    for resident in instance.single_residents:
        for pref in range(resident.num_preferences):
            encoding += "%d -%d 0" % (varmap[("r", resident.id, pref)],
                                      varmap[("r", resident.id, pref)]+1)
    # Clause 4
    for hospital in instance.single_hospitals:
        for pref in range(hospital.num_preferences):
            encoding += "%d -%d 0" % (varmap[("h", hospital.id, pref)],
                                      varmap[("h", hospital.id, pref)]+1)
    for hospital in instance.hospitals:
        for resident in instance.single_residents:
            # Clause 5
            # Note that position_of and rank_of are two different things.
            p = resident.position_of(hospital)
            q = hospital.position_of(resident)
            encoding += "-%d %d %d 0" % (varmap[("r", resident.id, p)],
                                         varmap[("r", resident.id, p)]+1,
                                         varmap[("h", hospital.id, q)])
            encoding += "-%d %d -%d 0" % (varmap[("r", resident.id, p)],
                                          varmap[("r", resident.id, p)]+1,
                                          varmap[("h", hospital.id, q)]+1)
            # Clause 6
            encoding += "-%d %d %d 0" % (varmap[("h", hospital.id, q)],
                                         varmap[("h", hospital.id, q)]+1,
                                         varmap[("r", resident.id, p)])
            encoding += "-%d %d -%d 0" % (varmap[("h", hospital.id, q)],
                                          varmap[("h", hospital.id, q)]+1,
                                          varmap[("r", resident.id, p)]+1)





"""Different ways of generating a random instance of HRTC."""


import random

from pyhrtc.basics import Agent, Couple, Instance

SCORE_MU = 80
SCORE_SIGMA = 5


def make_master_list(options, rounding=None):
    """Given a set of options, make a master list from the set. This will be
    done by assigning scores to each of the options, using the rounding
    function (if supplied) to round the scores, and finally sorting the options
    based on this score.
    """
    scores = {}
    for option in options:
        score = random.normalvariate(SCORE_MU, SCORE_SIGMA)
        if rounding:
            score = rounding(score)
        scores[option] = score
    return sorted(options, key=lambda x: scores[x])


def gen_capacities(total, hospitals, even_posts):
    """Generate capacities for a number of hospitals
    :param int hospitals: The number of hospitals
    :param int total: The number of posts to generate
    :param bool even_posts: are the posts distributed evenly (True) or randomly
                            (False)

    :return: the capacities of hospitals as a list
    :rtype: List
    """
    if even_posts:
        share = int(total / hospitals)
        excess = total - share*hospitals
        capacities = [share] * hospitals
        for hosp in range(excess):
            capacities[hosp] += 1
    else:
        capacities = [1] * hospitals
        for _ in range(total-hospitals):
            capacities[random.randrange(hospitals)] += 1
    return capacities


def random_hrtc(number_of_hospitals, number_of_single_residents,
                number_of_couples=0, resident_pref_length=None,
                hospital_pref_length=None, capacity=None, even_posts=False,
                resident_tie_density=0, hospital_tie_density=0,
                master_list=False, start_at_one=False):
    """Generates a random instance of HRTC, with the given properties.
    Note that couples are generated by interleaving two single residents.

    :param int number_of_hospitals: the number of hospitals
    :param int number_of_single_residents: the number of single residents
    :param int number_of_couples: the number of couples
    :param int resident_pref_length: how many preferences does each resident
                                     have
    :param int hospital_pref_length: how many preferences does each hospital
                                     have
    :param float resident_tie_density: the tie density of the residents
    :param float hospital_tie_density: the tie density of the hospitals
    :param bool master_list: are hospital preferences decided by a master list?
    :param int capacity: how many posts in total are there
    :param bool even_posts: are the posts distributed evenly (True) or randomly
    (False) :param bool start_at_one: do agent IDs have to start at 1. Needed
    for some solvers.

    """
    if resident_pref_length and hospital_pref_length:
        # I don't have a good way of doing this.
        raise NotImplementedError
    if not capacity:
        capacity = number_of_single_residents + 2 * number_of_couples
    capacities = gen_capacities(capacity, number_of_hospitals, even_posts)
    # Generate all the hospitals and agents
    hospitals = {ident + start_at_one: Agent(ident + start_at_one,
                                                capacities[ident]) for ident in
                 range(number_of_hospitals)}
    single_residents = {ident + start_at_one: Agent(ident + start_at_one)
                        for ident in range(number_of_single_residents)}
    couple_doctors = {}
    for ident in range(number_of_couples):
        offset = number_of_single_residents + 2 * ident + start_at_one
        couple_doctors[2*ident+start_at_one] = Agent(offset)
        couple_doctors[2*ident+1+start_at_one] = Agent(offset+1)
    # Generate a master list
    if master_list:
        master_list = [doctor.ident for doctor in single_residents.values()]
        master_list.extend(doctor.ident for doctor in couple_doctors.values())
        master_list = make_master_list(master_list)
    if not hospital_pref_length:
        for doctor in single_residents.values():
            doctor.make_random_preferences(hospitals.keys(),
                                           length=resident_pref_length,
                                           tie_density=resident_tie_density)
        for doctor in couple_doctors.values():
            doctor.make_random_preferences(hospitals.keys(),
                                           length=resident_pref_length,
                                           tie_density=resident_tie_density)
        for hospital in hospitals.values():
            options = [doctor.ident for doctor in single_residents.values()
                       if doctor.is_acceptable(hospital.ident)]
            options.extend([doctor.ident for doctor in couple_doctors.values()
                            if doctor.is_acceptable(hospital.ident)])
            if master_list:
                hospital.make_master_list_preferences(options, master_list)
            else:
                hospital.make_random_preferences(options,
                                                 tie_density=hospital_tie_density)
    else:
        # Hospital preference list lengths are not implemented
        raise NotImplementedError
    couples = {}
    for ident in range(number_of_couples):
        offset = 2 * ident + start_at_one
        couple = Couple.from_two_doctors(couple_doctors[offset],
                                         couple_doctors[offset + 1])
        couples[couple.ident] = couple
    instance = Instance(single_agents_left=single_residents, couples_left=couples,
                        single_agents_right=hospitals)
    return instance



"""Read, write and inspect instances of HRTC (and specialisations like SMTI,
HRT etc.)
"""


import csv
import re

from pyhrtc.basics import Agent, Couple, Instance
from pyhrtc.weightedinstance import WeightedAgent, WeightedInstance

JUST_NUMBER_RE = re.compile(r'^\d+$')


class UnknownFormatException(Exception):
    """An unknown format for the instance file."""

    def __init__(self):
        super().__init__("pyhrtc either does not support or "
                         "cannot read this file format.")


INSTANCE_READERS = {}
"""This dictionary should contain functions that, for a given variant of
instance format, takes as input a filename and returns an Instance of that
problem.
"""


def read_hrtc_glasgow_hrtc_nocolon(filename):
    """Reads in an instance of HRTC from filename in the usual Glasgow format,
    but with no colons.
    """
    with open(filename, "r") as infile:
        num_single_residents = int(infile.readline().rstrip())
        num_couples = int(infile.readline().rstrip())
        num_hospitals = int(infile.readline().rstrip())
        instance = Instance()
        for _ in range(num_single_residents):
            line = infile.readline()
            ident = int(line.split()[0])
            doc = Agent(ident)
            doc.read_preferences(line.split()[1:])
            instance.add_agent_left(doc)
        for _ in range(num_couples):
            line = infile.readline()
            id1 = int(line.split()[0])
            id2 = int(line.split()[1])
            couple = Couple(id1, id2)
            couple.read_preferences(line.split()[2:])
            instance.add_couple_left(couple)
        for _ in range(num_hospitals):
            line = infile.readline()
            ident = int(line.split()[0])
            cap = int(line.split()[1])
            hospital = Agent(ident, capacity=cap)
            hospital.read_preferences(line.split()[2:])
            instance.add_agent_right(hospital)
    return instance


def read_hrt_glasgow_nocolon(filename):
    """Reads in an instance of HRT from filename in the usual Glasgow format,
    except that the first line is a zero, the number of (single-only) residents
    is on the second line and the number of hospitals is on the third.
    """
    with open(filename, "r") as infile:
        infile.readline()  # First line is just 0
        num_doctor = int(infile.readline().rstrip())
        num_hospital = int(infile.readline().rstrip())
        instance = Instance()
        for _ in range(num_doctor):
            line = infile.readline()
            ident = int(line.split()[0])
            doctor = Agent(ident)
            doctor.read_preferences(line.split()[1:])
            instance.add_agent_left(doctor)
        for _ in range(num_hospital):
            line = infile.readline()
            ident = int(line.split()[0])
            hospital = Agent(ident, capacity=int(line.split()[1]))
            hospital.read_preferences(line.split()[2:])
            instance.add_agent_right(hospital)
    return instance


def read_iain_instance(filename):
    """Reads in an instance of HRTC from filename in Iain's format.
    """
    with open(filename, "r") as infile:
        total_num_residents = int(infile.readline().rstrip())
        num_hospitals = int(infile.readline().rstrip())
        num_couples = int(infile.readline().rstrip())
        num_single_residents = total_num_residents - 2*num_couples
        infile.readline()  # number jobs
        infile.readline()  # min pref list
        infile.readline()  # max pref list
        infile.readline()  # isEventDist
        infile.readline()  # res Popularity
        infile.readline()  # hos Popularity
        infile.readline()  # empty line
        instance = Instance()
        for _ in range(num_couples):
            line = infile.readline().rstrip()
            line_b = infile.readline().rstrip()
            ident = int(line.split()[0])
            ident2 = int(line_b.split()[0])
            couple = Couple(ident, ident2)
            couple.read_individual_preferences(line.split()[1:],
                                               line_b.split()[1:])
            instance.add_couple_left(couple)
        for _ in range(num_single_residents):
            line = infile.readline()
            ident = int(line.split()[0])
            doc = Agent(ident)
            doc.read_preferences(line.split()[1:])
            instance.add_agent_left(doc)
        for _ in range(num_hospitals):
            line = infile.readline()
            ident = int(line.split()[0])
            hospital = Agent(ident, capacity=int(line.split()[1]))
            hospital.read_preferences(line.split()[2:])
            instance.add_agent_right(hospital)
    return instance


def read_smti_grp_table_no_header(filename):
    """Reads an SMTI-GRP instance from a CSV file. The first two rows indicate
    the number of rows (first row) and columns (second row) that are in this
    file, not including the first two rows.
    :param filename: The name of the file containing the instance
    :type filename: string
    :return: the instance
    :rtype: WeightedInstance
    """
    lefts = {}
    rights = {}
    with open(filename, "r") as infile:
        int(infile.readline())
        columns = int(infile.readline())
        for left_id in range(1, columns+1):
            left_id = int(left_id)
            lefts[left_id] = WeightedAgent(left_id)
        right_id = 0
        for line in infile:
            right_id += 1
            right = WeightedAgent(right_id)
            left_id = 0
            for weight in line.split():
                left_id += 1
                weight = float(weight)
                right.add_weight(left_id, weight)
                lefts[left_id].add_weight(right.ident, weight)
            rights[right_id] = right
    return WeightedInstance(lefts, rights)


def read_smti_grp_table(filename):
    """Reads an SMTI-GRP instance from a CSV file. The first row and column are
    expected to be identifiers.
    :param filename: The name of the file containing the instance
    :type filename: string
    :return: the instance
    :rtype: WeightedInstance
    """
    lefts = {}
    rights = {}
    with open(filename, "r") as infile:
        reader = csv.DictReader(infile)
        topleft_header = reader.fieldnames[0]
        for left_id in reader.fieldnames[1:]:
            left_id = int(left_id)
            lefts[left_id] = WeightedAgent(left_id)
        for row in reader:
            right_id = int(row[topleft_header])
            right = WeightedAgent(right_id)
            for left_id, weight in row.items():
                if left_id == topleft_header:
                    continue
                left_id = int(left_id)
                weight = float(weight)
                right.add_weight(left_id, weight)
                lefts[left_id].add_weight(right.ident, weight)
            rights[right_id] = right
    return WeightedInstance(lefts, rights)


# Register the instance reader
INSTANCE_READERS["Glasgow_HRTC_nocolon"] = read_hrtc_glasgow_hrtc_nocolon
INSTANCE_READERS["Iain"] = read_iain_instance
INSTANCE_READERS["Glasgow_HRT_extraline"] = read_hrt_glasgow_nocolon
INSTANCE_READERS["SMTI-GRP Table"] = read_smti_grp_table
INSTANCE_READERS["SMTI-GRP Table no header"] = read_smti_grp_table_no_header


def write_hrtc_glasgow_hrtc_nocolon(instance, filename):
    """Writes an Instance to a file in the Glasgow HRTC, without colons.
    """
    write_hrtc_glasgow_hrtc(instance, filename, False)


def write_hrtc_glasgow_hrtc_colon(instance, filename):
    """Writes an Instance to a file in the Glasgow HRTC, with colons.
    """
    write_hrtc_glasgow_hrtc(instance, filename, True)


def write_hrtc_glasgow_hrtc(instance, filename, colon):
    """Writes an Instance to a file in the Glasgow HRTC, either with or without
    colons.
    """
    if colon:
        colon = ":"
    else:
        colon = ""
    with open(filename, "w") as outfile:
        outfile.write("%d\n" % instance.number_of_single_agents_left())
        outfile.write("%d\n" % instance.number_of_couples_left())
        outfile.write("%d\n" % instance.number_of_single_agents_right())
        for agent in instance.single_agents_left:
            outfile.write("%s%s %s\n" % (agent.ident, colon,
                                         agent.preference_string()))
        for couple in instance.couples_left:
            outfile.write("%s%s %s\n" % (couple.split_ident(), colon,
                                         couple.preference_string()))
        for agent in instance.single_agents_right:
            outfile.write("%s%s %d%s %s\n" % (agent.ident, colon,
                                              agent.capacity, colon,
                                              agent.preference_string()))


def write_hrtc_edin_hrtc(instance, filename):
    """Writes an Instance to a file in the Edinburgh? HRTC format
    """
    with open(filename, "w") as outfile:
        outfile.write("0\n")
        outfile.write("%d\n" % instance.number_of_single_agents_left())
        outfile.write("%d\n" % instance.number_of_single_agents_right())
        for agent in instance.single_agents_left:
            outfile.write("%s %s\n" % (agent.ident, agent.preference_string()))
        for agent in instance.single_agents_right:
            outfile.write("%s %d %s\n" % (agent.ident, agent.capacity,
                                          agent.preference_string()))


Instance.add_writer("Glasgow_HRTC_nocolon", write_hrtc_glasgow_hrtc_nocolon)
Instance.add_writer("Glasgow_HRTC_colon", write_hrtc_glasgow_hrtc_colon)
Instance.add_writer("Edin_HRTC", write_hrtc_edin_hrtc)


def read_hrtc_filetype(filename):
    """Given the name of a file containing an instance of HRTC, tries to
    determine the variant.

    :param filename: The name of the file to read
    :return: the variant as a string
    """
    with open(filename, "r") as infile:
        firstline = infile.readline().rstrip()
        variant = 0
        try:
            if "," in firstline:
                variant = "SMTI-GRP Table"
            elif int(firstline) == 0:
                variant = "Glasgow_HRT_extraline"
            else:
                try:
                    second_line = int(infile.readline().rstrip())
                except ValueError:
                    second_line = None
                third_line = infile.readline().rstrip()
                if ":" in third_line:
                    variant = "Glasgow_HRT_colon"
                elif JUST_NUMBER_RE.match(third_line):
                    fourth_line = infile.readline().rstrip()
                    if ":" in fourth_line:
                        variant = "Glasgow_HRTC_colon"
                    else:
                        infile.readline()  # 5th line
                        infile.readline()  # 6th line
                        seventh_line = infile.readline().rstrip()
                        if "false" in seventh_line or "true" in seventh_line:
                            variant = "Iain"
                        else:
                            variant = "Glasgow_HRTC_nocolon"
                elif second_line and third_line.count(" ") == second_line - 1:
                    variant = "SMTI-GRP Table no header"
        except ValueError:
            variant = firstline
    return variant


def read_hrtc(filename, variant=None):
    """Reads an instance of HRTC from the given file and returns the resulting
    Instance object.

    :param filename: The name of the file to read
    :param variant: The type of the file. If not present, will be autodetected.
    :return: An instance of HRTC
    """
    if variant is None:
        variant = read_hrtc_filetype(filename)
    if variant in INSTANCE_READERS:
        return INSTANCE_READERS[variant](filename)
    raise UnknownFormatException



"""Various IP models for HRTC and variants."""

from pulp import LpProblem, LpMaximize, LpVariable, lpSum, LpStatusOptimal

from pyhrtc.weightedinstance import WeightedInstance


class MAX_SMTI_IP(LpProblem):
    """An IP model to find a maximum cardinality stable matching.
    """

    def __init__(self, instance, name=None):
        if instance.number_of_couples_left() != 0:
            raise Exception("This model does not support couples")
        if name is None:
            name = "SMTI"
        super().__init__(name, LpMaximize)
        self._instance = instance
        self._built = False
        self._cardinality_restriction = None

        """Map from each variable name (as a string) to the actual variable object.
        _variables_lr is a map such that _variables_lr[left][right] gives the
        variable where Agent left is assigned to Agent right
        _variables_rl is a map such that _variables_rl[right][left] gives the
        variable where Agent right is assigned to Agent left
        """
        self._variables_lr = None
        self._variables_rl = None

        self._use_dummy = False
        """Maps the y variables, the dummy binary variables from Chapter 5 of paper.
        """
        self._variables_yl = None
        self._variables_yr = None

        """Should we find a maximum weight matching, or maximum size?"""
        self._weighted = False

    @property
    def dummy_variables(self):
        """Are dummy variables used in this model."""
        return self._use_dummy

    @dummy_variables.setter
    def dummy_variables(self, choice):
        """Enable (or disable) the use of dummy variables. Note that this only
        makes sense before building or solving.
        """
        if self._built:
            raise Exception("Cannot modify model after it is built")
        self._use_dummy = choice

    @property
    def weighted(self):
        """Are we finding a maximum weight matching"""
        return self._weighted

    @weighted.setter
    def weighted(self, choice):
        """Enable (or disable) the search for a maximum weight stable matching.
        """
        if self._built:
            raise Exception("Cannot modify model after it is built")
        if choice and not isinstance(self._instance, WeightedInstance):
            raise Exception("Cannot find a maximum weight stable matching if "
                            "the instance is not weighted")
        self._weighted = choice

    @property
    def cardinality_restriction(self):
        """Are we looking for a matching of a specific, known size?
        """
        return self._cardinality_restriction

    @cardinality_restriction.setter
    def cardinality_restriction(self, value):
        """Add a constraint to only search for matchings of the given size.
        """
        self._cardinality_restriction = value

    def _make_variables(self):
        """Makes the variables
        """
        if self._variables_lr is not None:
            return
        self._variables_lr = {}
        self._variables_rl = {}
        for left in self._instance.single_agents_left:
            self._variables_lr[left.ident] = {}
            for right in left.acceptable_agents():
                name = f"x_{left.ident}_{right}"
                variable = LpVariable(name, cat='Binary')
                self._variables_lr[left.ident][right] = variable
                if right not in self._variables_rl:
                    self._variables_rl[right] = {}
                self._variables_rl[right][left.ident] = variable

    def _make_capacity_constraints(self):
        """Makes the constraints for capacities.
        """
        for left in self._instance.single_agents_left:
            self += (lpSum(self._variables_lr[left.ident].values()) <= 1,
                     f"left_{left.ident}_capacity")
        for right in self._instance.single_agents_right:
            # If right.ident isn't in here, then this right agent has no
            # preferences.
            if right.ident not in self._variables_rl:
                continue
            self += (lpSum(self._variables_rl[right.ident].values()) <= 1,
                     f"right_{right.ident}_capacity")

    def _make_dummy_stability_constraints(self):
        """Makes a variable variable for each agent and each rank k in their
        preferences, which will be 1 if and only if the agent is matched with
        another agent of rank k or less.
        """
        if self._variables_yl is not None:
            return
        self._variables_yl = {}
        self._variables_yr = {}
        for left in self._instance.single_agents_left:
            self._variables_yl[left.ident] = {}
            name = f"y_l_{left.ident}_1"
            self._variables_yl[left.ident][1] = LpVariable(name, cat='Binary')
            sum_left = []
            for right_id in left.preferences[0]:
                sum_left.append(self._variables_lr[left.ident][right_id])
            self += (lpSum(sum_left) == self._variables_yl[left.ident][1])
            for k, tie_group in enumerate(left.preferences[1:], start=2):
                name = f"y_l_{left.ident}_{k}"
                self._variables_yl[left.ident][k] = LpVariable(name,
                                                               cat='Binary')
                sum_left = []
                for right_id in tie_group:
                    sum_left.append(self._variables_lr[left.ident][right_id])
                self += (lpSum(sum_left) + self._variables_yl[left.ident][k-1]
                         == self._variables_yl[left.ident][k])
        for right in self._instance.single_agents_right:
            self._variables_yr[right.ident] = {}
            name = f"y_r_{right.ident}_1"
            self._variables_yr[right.ident][1] = LpVariable(name, cat='Binary')
            sum_right = []
            for left_id in right.preferences[0]:
                sum_right.append(self._variables_lr[left_id][right.ident])
            self += (lpSum(sum_right) == self._variables_yr[right.ident][1])
            for k, tie_group in enumerate(right.preferences[1:], start=2):
                name = f"y_r_{right.ident}_{k}"
                self._variables_yr[right.ident][k] = LpVariable(name,
                                                                cat='Binary')
                sum_right = []
                for left_id in tie_group:
                    sum_right.append(self._variables_lr[left_id][right.ident])
                self += (lpSum(sum_right)
                         + self._variables_yr[right.ident][k-1]
                         == self._variables_yr[right.ident][k])
        for left in self._instance.single_agents_left:
            for k, tie_group in enumerate(left.preferences[1:], start=2):
                for right_id in tie_group:
                    right = self._instance.single_agent_right(right_id)
                    rank = right.rank_of(left.ident)
                    self += (1 - self._variables_yl[left.ident][k]
                             <= self._variables_yr[right_id][rank])

    def _make_stability_constraints(self):
        """Makes the stability constraints.
        """
        for left in self._instance.single_agents_left:
            for right_id in left.acceptable_agents():
                right = self._instance.single_agent_right(right_id)
                xiq = lpSum(self._variables_lr[left.ident][other]
                            for other in left.as_good_as(right.ident))
                xpj = lpSum(self._variables_rl[right.ident][other]
                            for other in right.as_good_as(left.ident))
                self += 1 - xiq <= xpj, f"{left.ident}.{right.ident}.stability"

    def _add_dummy_card_constraint(self):
        """Make a cardinality constraint obased on the dummy variables.
        """
        constraint = []
        for left in self._instance.single_agents_left:
            max_group = len(left.preferences)
            constraint.append(self._variables_yl[left.ident][max_group])
        self += (lpSum(constraint) == self._cardinality_restriction,
                 f"card-constraint")

    def _make_dummy_card_objective(self):
        """Make an objective based on the xij variables.
        """
        objective = []
        for left in self._instance.single_agents_left:
            max_group = len(left.preferences)
            objective.append(self._variables_yl[left.ident][max_group])
        self += lpSum(objective)

    def _make_card_objective(self):
        """Make an objective based on the xij variables.
        """
        objective = []
        for left in self._instance.single_agents_left:
            for right in left.acceptable_agents():
                if self.weighted:
                    objective.append(left.weight_of(right) *
                                     self._variables_lr[left.ident][right])
                else:
                    objective.append(self._variables_lr[left.ident][right])
                objective.append(self._variables_lr[left.ident][right])
        self += lpSum(objective)

    def build_model(self):
        """Build the model.
        """
        if self._built:
            return
        self._make_variables()
        self._make_capacity_constraints()
        if self._use_dummy:
            self._make_dummy_stability_constraints()
            if not self.weighted:
                self._make_dummy_card_objective()
            else:
                if self._cardinality_restriction is not None:
                    self._add_dummy_card_constraint()
                self._make_regular_objective()
        else:
            self._make_stability_constraints()
            self._make_card_objective()
        self._built = True

    def solve(self):
        """Solves the problem, and returns the matching. The matching is a map
        such that matching[left] = right if and only if left is matched to
        right.
        """
        self.build_model()
        super().solve()
        if self.status != LpStatusOptimal:
            raise Exception("Not optimal solution.")
        matching = {}
        for left in self._instance.single_agents_left:
            for right in left.acceptable_agents():
                # Because some solvers sometimes return "close to 1" instead of
                # 1, even for binary values, we just get close.
                if self._variables_lr[left.ident][right].varValue >= 0.95:
                    matching[left.ident] = right
        return matching



"""Basic elements of HRTC problems."""


import random


def grouped(things):
    """Given some iterable object, returns an iterator that goes through the
    things in the object two at a time.
    """
    return zip(*[iter(things)]*2)


class Agent():
    """An agent."""

    def __init__(self, ident, capacity=1):
        self._ident = ident
        self._capacity = capacity
        # This will be a list of lists, with each inner list corresponding to a
        # tie group
        self._preferences = []
        self._num_preferences = None

    def __str__(self):
        """A human readable string representation of this Agent.
        """
        return (f"Agent %d with preferences: %s" % (self._ident,
                                                    self.preference_string()))

    def __repr__(self):
        return f"Agent:%d" % self._ident

    @property
    def ident(self):
        """The ID of this agent."""
        return self._ident

    @ident.setter
    def ident(self, new):
        """We cannot change the ID of an agent"""
        raise NotImplementedError

    @property
    def capacity(self):
        """How many other agents can this agent can support."""
        return self._capacity

    @capacity.setter
    def capacity(self, new):
        """Not supported."""
        raise NotImplementedError

    @property
    def num_preferences(self):
        """The number of preferences of this agent, aka the length of its
        preference list.
        """
        if self._num_preferences is None:
            self._num_preferences = sum([len(x) for x in self.preferences])
        return self._num_preferences

    @property
    def preferences(self):
        """Return the preferences, as a list of tie groups. Entries which are
        not in a tie at all still appear by themselvs in a list (i.e. 1 (2 3)
        is [[1], [2, 3]]).
        """
        return self._preferences

    def is_empty(self):
        """Return True if and only if this Agent has no preferences, and
        therefore can be safely deleted.
        """
        return not self.preferences

    @preferences.setter
    def preferences(self, new):
        """Set the preferences of this agent. This must be a list of lists,
        such that each tie group is a list. For instance, 1 (2 3) should be
        passed in as [[1], [2, 3]].
        """
        self._preferences = new
        self._num_preferences = None

    def rank_of(self, other):
        """Find the rank of other according to this agent. Note that ranks
        start at 1, not 0.
        """
        if isinstance(other, Agent):
            ident = other.ident
        else:
            ident = other
        for index, group in enumerate(self.preferences, start=1):
            if ident in group:
                return index
        return -1

    def trim_after_worst(self, agents):
        """Given a set of agents, trim this agents preference list by removing
        anything that occurs after the worst agent in agents.
        :param agents: An Iterable containing agent IDs
        :return: The number of preferences removed
        """
        removed = 0
        while not [agent for agent in agents if agent in self.preferences[-1]]:
            removed += len(self.preferences.pop(-1))
        return removed

    def position_of(self, other):
        """Find the position of other in this preference list. Note that this
        is not the same as rank, this function assumes an ordering on the
        elements within a tie, and gives an absolute count on the number of
        agents before it.
        """
        count = 0
        for group in self.preferences:
            for item in group:
                if item == other.id:
                    return count
                count += 1
        return -1

    def tie_density(self):
        """Get the tie density according to this agent."""
        if len(self.preferences) == 1 and len(self.preferences[1]) == 1:
            raise NotImplementedError()
        count = 0
        for pref in self.preferences:
            count += len(pref) - 1
        return count / (count + len(self.preferences))

    def make_random_preferences(self, options, tie_density=0, length=None):
        """Create a random preference list from the list options, with the
        given tie density. If length is given, then stop assigning preferences
        once "length" items have been added.
        """
        if length:
            chosen = random.sample(options, length)
        else:
            chosen = list(options)  # Copy so we don't shuffle the original
        random.shuffle(chosen)
        # Reset preferences
        self._preferences = []
        current_tie = []
        for item in chosen:
            # If there is a tie, and we want to end it, do so;
            if current_tie and random.random() >= tie_density:
                # Start a new tie group
                self._preferences.append(current_tie)
                current_tie = []
            current_tie.append(item)
        # Don't forget to add the last tie
        self._preferences.append(current_tie)

    def make_master_list_preferences(self, options, master_list, length=None):
        """Create a preference list from the list options, with the given tie
        density and master list. If length is given, then stop assigning
        preferences once "length" items have been added.
        """
        chosen = []
        for item in master_list:
            if item in options:
                chosen.append(item)
                if length and len(chosen) == length:
                    break
        # Reset preferences
        self._preferences = []
        for item in chosen:
            self._preferences.append([item])

    def acceptable_agents(self):
        """An iterator for acceptable agents for this Agent."""
        for tie in self.preferences:
            for agent in tie:
                yield agent

    def as_good_as(self, other):
        """An iterator for all agent IDs that this agent thinks are at least as good as other.
        :param other: an Agent ID
        :return: An iterator of Agent IDs
        """
        for tie in self.preferences:
            for agent in tie:
                yield agent
            if other in tie:
                break

    def is_acceptable(self, ident):
        """Is the given agent (as identified by their identifier) acceptable
        for this Agent?
        """
        for tie in self.preferences:
            if ident in tie:
                return True
        return False

    def preference_string(self):
        """Returns the string of preferences for this agent."""
        def format_tie(tie_as_list):
            """Given a set of tied elements, returns a string representing
            them, by surrounding with brackets if there is more than one item.
            """
            if not tie_as_list:
                return ""
            if len(tie_as_list) == 1:
                return "%s" % tie_as_list[0]
            return "(%s)" % (" ".join(map(str, tie_as_list)))
        return " ".join([format_tie(tie) for tie in self.preferences])

    def read_preferences(self, tokens):
        """Read in and assign preferences based on the given string.
        """
        # First reset preferences
        self._preferences = []
        in_tie = False
        current = []
        for token in tokens:
            if not in_tie:
                # Tie starting
                if token[0] == "(" and token[-1] != ")":
                    in_tie = True
                    current.append(int(token[1:]))
                else:
                    # Not at all a tie
                    if token[0] == "(" and token[-1] == ")":
                        token = token[1:-1]
                    self._preferences.append([int(token)])
            else:
                # Inside a tie
                if token[-1] == ")":
                    # Tie ends here
                    current.append(int(token[:-1]))
                    self._preferences.append(current)
                    in_tie = False
                    current = []
                else:
                    # Tie keeps going
                    current.append(int(token))


class Couple(Agent):
    """Two agents together."""

    def __init__(self, first, second):
        super().__init__("(%s,%s)" % (first, second))
        self._first = first
        self._second = second

    def split_ident(self):
        """Returns a string representation of the two individual agents,
        rather than the combined couple.
        """
        return "%s %s" % (self._first, self._second)

    def read_individual_preferences(self, tokens_a, tokens_b):
        """Read in and assign preferences, when preferences are given in two
        distinct lists of strings.
        """
        self._preferences = []
        for one, two in zip(tokens_a, tokens_b):
            self._preferences.append([int(one), int(two)])

    def read_preferences(self, tokens):
        """Read in and assign preferences based on the given string.
        """
        # First reset preferences
        self._preferences = []
        in_tie = False
        current = []
        for token1, token2 in grouped(tokens):
            if not in_tie:
                # Tie starting
                if token1[0] == "(":
                    in_tie = True
                    pref = (int(token1[1:]), int(token2))
                    current.append(pref)
                else:
                    # Not at all a tie
                    self._preferences.append([(int(token1), int(token2))])
            else:
                # Inside a tie
                if token2[-1] == ")":
                    # Tie ends here
                    current.append((int(token1), int(token2[:-1])))
                    self._preferences.append(current)
                    in_tie = False
                    current = []
                else:
                    # Tie keeps going
                    current.append((int(token1), int(token2)))

    def preference_string(self):
        """Returns the string of preferences for this agent."""
        def format_tie(tie_as_list):
            """Given a set of tied elements, returns a string representing
            them, by surrounding with brackets if there is more than one item.
            """
            if not tie_as_list:
                return ""
            if len(tie_as_list) == 1:
                return "%s %s" % (tie_as_list[0][0], tie_as_list[0][1])
            return "(%s)" % (" ".join(["%s %s" % (c[0], c[1])
                                       for c in tie_as_list]))
        return " ".join([format_tie(tie) for tie in self.preferences])

    def __str__(self):
        """A human readable string representation of this Agent.
        """
        return (f"Agent (%d, %d) with preferences: %s" % (self._first,
                                                          self._second,
                                                          self.preference_string()))

    def __repr__(self):
        return f"Agent:%d,%d" % (self._first, self._second)

    @staticmethod
    def from_two_agents(agent1: Agent, agent2: Agent):
        """Given two agents, create a Couple from them.
        """
        couple = Couple(agent1.ident, agent2.ident)
        new_preferences = []
        for second_ind, second_pref in enumerate(agent2.preferences):
            for first_ind, first_pref in enumerate(agent1.preferences):
                if second_ind < first_ind:
                    continue
                if len(first_pref) != 1 or len(second_pref) != 1:
                    raise NotImplementedError("pyhrtc can't interleave "
                                              "from agents with ties")
                if first_ind == second_ind:
                    new_preferences.append([(first_pref[0], second_pref[0])])
                else:
                    first_alt = agent1.preferences[second_ind]
                    second_alt = agent2.preferences[first_ind]
                    new_preferences.append([(first_pref[0], second_pref[0]),
                                            (first_alt[0], second_alt[0])])
        couple.preferences = new_preferences
        return couple


class Instance():
    """An instance of HRTC.
    """

    """A dict to hold all functions that can write Instances to a file.
    """
    instance_writers = {}

    @staticmethod
    def add_writer(name, function):
        """Adds a writer function.
        :param name: The name (identifier) of the file format this function
        writes
        :type name: string
        :param function: A function which takes two parameters, an Instance and
        a string, and writes the instance to file as named in the string
        :type function: a function name
        """
        Instance.instance_writers[name] = function

    def __init__(self, single_agents_left=None, couples_left=None,
                 single_agents_right=None):
        """Create an Instance. Note that if any of the parameters are passed
        in, they are used as is (i.e. not copied) so don't modify the dicts
        after creating an instance. We use "left" and "right" to refer to the
        two sides of this instance.

        :param dict single_agents_left: A dictionary of single agents on the
        left side of this instance
        :param dict couples_left: A dictionary of couples on the left side of
        this instance
        :param dict single_agents_right: A dictionary of single agents on the
        right side of this instance
        """

        # Each of these is a map from ID to the actual entity, so make sure
        # you set IDs appropriately
        if single_agents_left:
            self._single_agents_left = single_agents_left
        else:
            self._single_agents_left = {}
        if couples_left:
            self._couples_left = couples_left
        else:
            self._couples_left = {}
        if single_agents_right:
            self._single_agents_right = single_agents_right
        else:
            self._single_agents_right = {}

    def number_of_single_agents_left(self) -> int:
        """The number of single residents on the left."""
        return len(self._single_agents_left)

    def number_of_couples_left(self) -> int:
        """The number of couples on the left."""
        return len(self._couples_left)

    def number_of_single_agents_right(self) -> int:
        """The number of single agents on the right."""
        return len(self._single_agents_right)

    @property
    def single_agents_left(self):
        """Returns a list of all single agents on the left."""
        return list(self._single_agents_left.values())

    @single_agents_left.setter
    def single_agents_left(self, new):
        """Not allowed."""
        raise NotImplementedError

    @property
    def couples_left(self):
        """Returns a list of all the couples on the left."""
        return list(self._couples_left.values())

    @couples_left.setter
    def couples_left(self, new):
        """Not allowed."""
        raise NotImplementedError

    @property
    def single_agents_right(self):
        """Returns a list of all the single agents on the right."""
        return list(self._single_agents_right.values())

    @single_agents_right.setter
    def single_agents_right(self, new):
        """Not allowed."""
        raise NotImplementedError

    def single_agent_left(self, ident):
        """Returns the agent on the left identified by ident.
        :param ident: The ID of the desired agent.
        :type ident: integer
        :returns: the agent
        :rtype: Agent
        """
        return self._single_agents_left[ident]

    def single_agent_right(self, ident):
        """Returns the agent on the right identified by ident.
        :param ident: The ID of the desired agent.
        :type ident: integer
        :returns: the agent
        :rtype: Agent
        """
        return self._single_agents_right[ident]

    def add_agent_left(self, agent):
        """Add a single agent to the left side of this instance.
        """
        self._single_agents_left[agent.ident] = agent

    def add_couple_left(self, couple):
        """Adds a couple to the left side of this instance.
        """
        self._couples_left[couple.ident] = couple

    def add_agent_right(self, agent):
        """Adds an agent to the right side of this instance.
        """
        self._single_agents_right[agent.ident] = agent

    def make_couple_from_agent_pair_on_left(self, number=1):
        """Take "number x 2" agents from the left, and turn them into couples
        by interleaving. Note that there is no reliable way to select which
        agents.
        """
        while number:
            if len(self._single_agents_left) < 2:
                raise LookupError("Don't have enough agents left")
            id1, id2 = list(self._single_agents_left.keys())[-2:]
            first = self._single_agents_left[id1]
            second = self._single_agents_left[id2]
            couple = Couple.from_two_agents(first, second)
            self._couples_left[couple.ident] = couple
            del self._single_agents_left[id1]
            del self._single_agents_left[id2]
            number -= 1

    def is_SMTI(self):
        """Returns true if this is an instance of SMTI (aka there are no
        couples, and each agent on the right has capacity 1.
        """
        if self.number_of_couples_left() != 0:
            return False
        for agent in self.single_agents_right:
            if agent.capacity != 1:
                return False
        return True

    def write_to_file(self, filename, variant="Glasgow_HRTC_nocolon"):
        """Writes the instance to a file."""
        if variant in Instance.instance_writers:
            Instance.instance_writers[variant](self, filename)
        else:
            raise Exception("pyhrtc either does not support or "
                            "cannot read this file format.")

    def long_string(self):
        """Write a long human-readable description of this instance.
        """
        return ("\n".join([str(agent) for agent in self.single_agents_left]) +
                "\n" +
                "\n".join([str(agent) for agent in self.single_agents_right]) +
                "\n")

    def __str__(self):
        """A human readable string representation of this instance.
        """
        return (f"Instance with %d single agents on the left, %d couples on "
                "the left and %d single agents on the right" %
                (self.number_of_single_agents_left(),
                 self.number_of_couples_left(),
                 self.number_of_single_agents_right()))

    def preprocess(self, side='L', order=None):
        """Preprocess this instance, removing any pairs which will definitely
        not be part of a stable matching.
        For each agent l on the Left, we build up a set O containing elements
        from the Right.  Then we set T to be all those agents on the Left which
        any agent from O considers at least as good as l.  If the total
        capacity of O is >= the total capacity of T, then in a stable matching
        l will be matched with an agent that l thinks is no worse than the
        worst agent in O.
        :param side: From which side are we removing preferences, 'L' or 'R'
        :param order: An iterating function to determine how we build up the
        set positions. It should take 5 parameters: agent, instance, positions,
        candidates, and side.  Agent is an Agent object, the one whose list we
        are preprocessing. The default implementation is to simply iterate
        through the preferences of this agent in order. More complex ordering
        functions might look at either positions or candidates to see if
        yielding a particular agent will significantly increase the size of
        candidates, which may be less useful for preprocessing.  instance is
        this instance, which may be useful when designing these more complex
        orderings. Side is as for this function.
        :return: How many preferences have been removed. Note that if the pair
        (a,b) is now no longer considered, this counts as one removal, not two.
        """
        removed = 0
        if order is None:
            def order(agent, instance, positions, candidates, side):
                for group in agent.preferences:
                    for other_agent in group:
                        if other_agent not in positions:
                          yield other_agent
        if side == 'L':
            print("looking at doctors")
            these = self.single_agents_left
            other = self.single_agents_right
            these_agent = self.single_agent_left
            other_agent = self.single_agent_right
        else:
            print("looking at hospitals")
            these = self.single_agents_right
            other = self.single_agents_left
            these_agent = self.single_agent_right
            other_agent = self.single_agent_left
        for single in these:
            positions = set()
            candidates = set()
            # Go through agents in this preference list according to some
            # ordering.
            for each in order(single, self, positions, candidates, side):
                positions.add(each)
                # Add anything that each thinks is at least as good as single
                for group in other_agent(each).preferences:
                    for agent in group:
                        candidates.add(agent)
                    if single.ident in group:
                        break
                if (sum([other_agent(pos).capacity for pos in positions]) >=
                    sum([these_agent(cand).capacity for cand in candidates])):
                    removed += single.trim_after_worst(positions)
                    self.clean_one(single, other)
        return removed

    def clean_one(self, agent, other_side):
        """Clean this instance, removing incompatible agents. This variant
        knows which agent has had its preferences edited, so it only goes
        through one set of agents, and only looks to remove one particular
        agent from a preference group.
        """
        found = False
        for other in other_side:
            for group in other.preferences:
                if agent.ident in group and not agent.is_acceptable(other.ident):
                    group.remove(agent.ident)
                    found = True
                    break
            if found:
                other.preferences = [group for group in other.preferences if group]

    def clean(self):
        """Clean this instance, removing from preference lists any incompatible
        agents.
        """
        for queue, other_queue in [(self.single_agents_left, self.single_agent_right),
                                   (self.single_agents_right, self.single_agent_left)]:
            for agent in queue:
                new_preferences = []
                for group in agent.preferences:
                    new_group = [other for other in group
                                if other_queue(other).is_acceptable(agent.ident)]
                    if new_group:
                        new_preferences.append(new_group)
                agent.preferences = new_preferences


import pyhrtc.fileio



"""Contains some calls to algorithms from 3rd parties.
"""

from networkx import Graph as NxGraph, connected_component_subgraphs
from networkx.algorithms.bipartite import maximum_matching
from networkx.algorithms.components import is_connected
from networkx.algorithms.matching import max_weight_matching as nx_max_weight

from pyhrtc.weightedinstance import WeightedInstance


def max_card_matching(instance):
    """Given an instance, calculate the cardinality of the biggest matching.
    Note that this matching need not be, and probably won't be, stable.

    :param instance: The instance in question
    :type instance: Instance
    :rtype: int
    :return: The size of the largest cardinality matching.
    """
    graph = NxGraph()
    if instance.number_of_couples_left() != 0:
        raise Exception("max card matching does not currently support couples")
    for left in instance.single_agents_left:
        graph.add_node(f"l%d" % left.ident, bipartite=0)
    for right in instance.single_agents_right:
        for cap in range(right.capacity):
            graph.add_node(f"r%d_%d" % (right.ident, cap), bipartite=1)
    for left in instance.single_agents_left:
        for pref_group in left.preferences:
            for right_id in pref_group:
                for cap in range(instance.single_agent_right(right_id).capacity):
                    graph.add_edge(f"l%d" % left.ident,
                                   f"r%d_%d" % (right_id, cap))
    size = 0
    for component in connected_component_subgraphs(graph):
        size += len(maximum_matching(component))
    return int(size/2)


def max_weight_matching(instance):
    """Given a weighted instance, calculate the weight of of the maximum weight
    matching.  Note that this matching need not be, and probably won't be,
    stable.

    :param instance: The instance in question
    :type instance: WeightedInstance
    :rtype: float
    :return: The weight of the largest weigh tmatching.
    """
    if not isinstance(instance, WeightedInstance):
        raise Exception("Max weight matching needs a WeightedInstance")
    graph = NxGraph()
    if instance.number_of_couples_left() != 0:
        raise Exception("Max weight matching does not "
                        "currently support couples")
    for left in instance.single_agents_left:
        graph.add_node(f"l%d" % left.ident, bipartite=0)
    for right in instance.single_agents_right:
        for cap in range(right.capacity):
            graph.add_node(f"r%d_%d" % (right.ident, cap), bipartite=1)
    for left in instance.single_agents_left:
        for pref_group in left.preferences:
            for right_id in pref_group:
                for cap in range(instance.single_agent_right(right_id).capacity):
                    graph.add_edge(f"l%d" % left.ident,
                                   f"r%d_%d" % (right_id, cap),
                                   weight=left.weight_of(right_id))
    weight = 0
    for component in connected_component_subgraphs(graph):
        for start, end in nx_max_weight(component):
            weight += graph[start][end]['weight']
    return weight
